package org.semwidg.semwidgql.internal;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

import org.semwidg.semwidgql.Prefix;
import org.semwidg.semwidgql.internal.element.ConditionalOperator;
import org.semwidg.semwidgql.internal.element.Element;
import org.semwidg.semwidgql.internal.element.Expression;
import org.semwidg.semwidgql.internal.element.KeywordExpression_Aggregate;
import org.semwidg.semwidgql.internal.element.KeywordExpression_Hide;
import org.semwidg.semwidgql.internal.element.KeywordExpression_Language;
import org.semwidg.semwidgql.internal.element.KeywordExpression_Optional;
import org.semwidg.semwidgql.internal.element.KeywordExpression_Predicate;
import org.semwidg.semwidgql.internal.element.KeywordExpression_Self;
import org.semwidg.semwidgql.internal.element.KeywordExpression_Timeend;
import org.semwidg.semwidgql.internal.element.KeywordExpression_Timeinterval;
import org.semwidg.semwidgql.internal.element.KeywordExpression_Timestart;
import org.semwidg.semwidgql.internal.element.KeywordExpression_Type;
import org.semwidg.semwidgql.internal.element.PathExpression;
import org.semwidg.semwidgql.internal.element.Property;
import org.semwidg.semwidgql.internal.element.RelationalExpression;
import org.semwidg.semwidgql.internal.element.Resource;
import org.semwidg.semwidgql.internal.element.ResourceOrProperty;
import org.semwidg.semwidgql.internal.element.SimpleQuery;
import org.semwidg.semwidgql.internal.factoryHelper.GroupExpression;
import org.semwidg.semwidgql.internal.factoryHelper.HideExpression;

public class Factory {

	public static String unescape(String value) {

		// unescape \.
		value = value.replaceAll("/\\./g", ".");
		return value;
	};

	public static String buildQuery(SimpleQuery simpleQuery) {

		StringBuilder str = new StringBuilder("");

		addStr(str, buildPrefixPart(simpleQuery));
		addStr(str, buildSelectPart(simpleQuery));
		addStr(str, buildWherePart(simpleQuery));
		addStr(str, buildGroupPart(simpleQuery));

		if (str.toString().indexOf("a owl:Thing") > -1) {
			str = new StringBuilder("PREFIX owl: <http://www.w3.org/2002/07/owl#>\n" + str.toString());
		}

		return str.toString();

	};

	public static String buildPrefixPart(SimpleQuery simpleQuery) {

		StringBuilder str = new StringBuilder("");

		if (simpleQuery != null) {

			ArrayList<String> prefixes = Sparql.getUsedPrefixStrings(simpleQuery.getPointer(), null);
			ArrayList<Prefix> usedPrefixObjects = Sparql.getUsedPrefixObjects(prefixes);

			if (usedPrefixObjects != null && usedPrefixObjects.size() > 0) {
				for (int i = 0; i < usedPrefixObjects.size(); i++) {
					addStr(str, "PREFIX ");
					addStr(str, usedPrefixObjects.get(i).getPrefix());
					addStr(str, ":");
					addStr(str, " <" + usedPrefixObjects.get(i).getUrl() + ">");
					addStr(str, "\n");
				}
			}
		}

		return unescape(str.toString());
	};

	public static String buildSelectPart(SimpleQuery simpleQuery) {

		StringBuilder str = new StringBuilder("");

		if (simpleQuery != null) {

			ArrayList<String> hiddenElements = getHiddenElements(simpleQuery);

			// If group expressions exist, results must be aggregated
			ArrayList<GroupExpression> groupExpressions = getGroupExpressions(simpleQuery);
			boolean aggregate = groupExpressions.size() > 0;

			ArrayList<ResourceOrProperty> resourcesAndProperties = Sparql.getUsedResourcesAndProperties(simpleQuery);

			addStr(str, "SELECT DISTINCT ");

			String aggregateFunction = null;

			for (int i = 0; i < resourcesAndProperties.size(); i++) {

				ResourceOrProperty elem = resourcesAndProperties.get(i);
				String variableName = Sparql.convertQueryElementToVariableName(elem);

				boolean hidden = hiddenElements.indexOf(variableName) >= 0;
				if (hidden) {
					continue;
				}

				aggregateFunction = getAggregateExpressions(elem, aggregate);

				String prediacte = getPredicate(elem);

				if (elem instanceof Resource) {

					if (((Resource) elem).getValue().equals("*")) {
						// wildcard
						if (aggregateFunction != null) {
							addStr(str, aggregateFunction + "(");
						}
						addStr(str, variableName);
						if (aggregateFunction != null) {
							addStr(str, ") AS " + variableName);
						}
						addStr(str, " ");
					}else {
						// normal resource
						if (aggregateFunction != null) {
							addStr(str, aggregateFunction + "(");
						}
						addStr(str, "(" + ((Resource) elem).getValue() + ")");
						if (aggregateFunction != null) {
							addStr(str, ")");
						}
						addStr(str, " AS " + variableName + " ");
					}

				}else {

					if (prediacte != null) {
						addStr(str, prediacte + " ");
					}

					if (aggregateFunction != null) {
						addStr(str, aggregateFunction + "(");
					}
					addStr(str, variableName);
					if (aggregateFunction != null) {
						addStr(str, ") AS " + variableName);
					}
					addStr(str, " ");
				}

			}

			addStr(str, "\n");
		}

		return unescape(str.toString());

	};

	public static ArrayList<String> getHiddenElements(SimpleQuery simpleQuery) {

		ArrayList<String> hiddenElements = new ArrayList<String>();
		ArrayList<HideExpression> hideExpressions = getHideExpressions(simpleQuery);
		for (int i = 0; i < hideExpressions.size(); i++) {
			if (hideExpressions.get(i).getFilter().getHide().toLowerCase().equals("true")) {
				hiddenElements.add(Sparql.convertQueryElementToVariableName(
						hideExpressions.get(i).getElement()));
			}
		}
		return hiddenElements;
	};

	public static ArrayList<HideExpression> getHideExpressions(Element elem) {

		ArrayList<HideExpression> arr = new ArrayList<HideExpression>();

		if (elem == null) {
			return arr;
		}

		int i = 0;

		if (elem instanceof SimpleQuery) {
			arr.addAll(getHideExpressions(((SimpleQuery) elem).getPathExpression()));
		}
		else if (elem instanceof PathExpression) {
			if (((PathExpression) elem).getMultiplePathExpression() != null) {
				for (i = 0; i < ((PathExpression) elem).getMultiplePathExpression().getPathExpressions().size(); i++) {
					arr.addAll(getHideExpressions(((PathExpression) elem).getMultiplePathExpression().getPathExpressions().get(i)));
				}
			}

			if (((PathExpression) elem).getProperties() != null) {
				for (i = 0; i < ((PathExpression) elem).getProperties().size(); i++) {
					arr.addAll(getHideExpressions(((PathExpression) elem).getProperties().get(i)));
				}
			}
		}
		else if (elem instanceof Property) {
			if (((Property) elem).getFilter() != null) {
				for (i = 0; i < ((Property) elem).getFilter().size(); i++) {

					Element filter = ((Property) elem).getFilter().get(i);
					if (filter instanceof KeywordExpression_Hide) {
						arr.add(new HideExpression((Property) elem, (KeywordExpression_Hide) filter));
					}
				}
			}
		}

		return arr;
	};

	public static String getAggregateExpressions(ResourceOrProperty elem, boolean forceAggregateExpression) {

		if (elem != null && elem.getFilter() != null) {

			ArrayList<Element> filterExpressions = elem.getFilter();

			for (int i = 0; i < filterExpressions.size(); i++) {
				if (filterExpressions.get(i) instanceof KeywordExpression_Aggregate) {
					return (((KeywordExpression_Aggregate) filterExpressions.get(i)).getAggregate()).trim().toUpperCase();
				}
			}
		}

		if (forceAggregateExpression) {
			return "SAMPLE";
		} else {
			return null;
		}
	};

	public static String getPredicate(ResourceOrProperty elem) {

		if (elem != null && elem.getFilter() != null) {

			ArrayList<Element> filterExpressions = elem.getFilter();

			for (int i = 0; i < filterExpressions.size(); i++) {
				if (filterExpressions.get(i) instanceof KeywordExpression_Predicate 
						&& ((KeywordExpression_Predicate) filterExpressions.get(i)).getPredicate().toLowerCase().equals("true")) {

					String value = elem.getValue();
					if (value.equals("*")) {
						value = Sparql.convertQueryElementToVariableName(elem);
					}

					String type = elem.getType();

					switch(type) {
					case "ForwardProperty":
					case "ForwardPropertyWithFilter":
					case "ReversedProperty":
					case "ReversedPropertyWithFilter":
						return "('" + value + "') AS ?PredicateFor" + Sparql.convertQueryElementToVariableName(elem).substring(1);
					case "ForwardWildcard":
					case "ForwardWildcardWithFilter":
					case "ReversedWildcard":
					case "ReversedWildcardWithFilter":
						return "?PredicateFor" + Sparql.convertQueryElementToVariableName(elem).substring(1);
					}
				}
			}
		}

		return null;
	};

	public static boolean getOptional(ResourceOrProperty elem) {

		if (elem != null && elem.getFilter() != null) {

			ArrayList<Element> filterExpressions = elem.getFilter();

			for (int i = 0; i < filterExpressions.size(); i++) {
				if (filterExpressions.get(i) instanceof KeywordExpression_Optional) {
					return ((KeywordExpression_Optional) filterExpressions.get(i)).getOptional().trim().toLowerCase().equals("true");
				}
			}
		}

		return false;
	};

	public static String buildWherePart(SimpleQuery simpleQuery) {

		StringBuilder str = new StringBuilder("");

		if (simpleQuery != null) {

			addStr(str, "WHERE {\n");

			addStr(str, buildPathExpression(simpleQuery.getResource(), simpleQuery.getPathExpression()));

			addStr(str, "}");

		}

		return unescape(str.toString());

	};

	public static String buildPathExpression(ResourceOrProperty previousElement, PathExpression pathExpression) {

		StringBuilder str = new StringBuilder("");
		boolean optional = false;

		if (pathExpression != null) {

			if (pathExpression.getProperties() != null && pathExpression.getProperties().size() > 0) {

				Property prop = pathExpression.getProperties().get(0);

				if (previousElement instanceof Resource) {

					optional = Factory.getOptional(prop);
					if (optional) {
						addStr(str, "\tOPTIONAL { ##" + Sparql.convertQueryElementToVariableName(prop) + " \n");
					}

					addStr(str, buildFirstTriple(previousElement, prop));
					addStr(str, buildFilter(previousElement));

					if (optional) {
						addStr(str, "\t} ##" + Sparql.convertQueryElementToVariableName(prop) + " \n");
					}

				}else {

					optional = Factory.getOptional(prop);
					if (optional) {
						addStr(str, "\tOPTIONAL { ##" + Sparql.convertQueryElementToVariableName(prop) + " \n");
					}

					addStr(str, Factory.buildTriple(previousElement, prop));

					if (optional) {
						addStr(str, "\t} ##" + Sparql.convertQueryElementToVariableName(prop) + " \n");
					}
				}

				addStr(str, Factory.buildFilter(prop));

				for (int i = 0; i < pathExpression.getProperties().size() - 1; i++) {

					optional = getOptional(pathExpression.getProperties().get(i + 1));
					if (optional) {
						addStr(str, "\tOPTIONAL { ##" + Sparql.convertQueryElementToVariableName(pathExpression.getProperties().get(i + 1)) + " \n");
					}

					addStr(str, buildTriple(pathExpression.getProperties().get(i), pathExpression.getProperties().get(i + 1)));
					addStr(str, buildFilter(pathExpression.getProperties().get(i + 1)));

					if (optional) {
						addStr(str, "\t} ##" + Sparql.convertQueryElementToVariableName(pathExpression.getProperties().get(i + 1)) + " \n");
					}
				}

				if (pathExpression.getMultiplePathExpression() != null && pathExpression.getMultiplePathExpression().getPathExpressions().size() > 0) {
					for (int j = 0; j < pathExpression.getMultiplePathExpression().getPathExpressions().size(); j ++) {

						PathExpression pathEx = pathExpression.getMultiplePathExpression().getPathExpressions().get(j);

						Property property = pathExpression.getProperties().get(pathExpression.getProperties().size() - 1);
						// addStr(str, "\tOPTIONAL { \n";
						addStr(str, buildPathExpression(property, pathEx));
						// addStr(str, "\t} \n";
					}
				}

			}else if (pathExpression.getMultiplePathExpression() != null && pathExpression.getMultiplePathExpression().getPathExpressions().size() > 0){
				for (int k = 0; k < pathExpression.getMultiplePathExpression().getPathExpressions().size(); k ++) {

					PathExpression pathEx = pathExpression.getMultiplePathExpression().getPathExpressions().get(k);
					// addStr(str, "\tOPTIONAL { \n";
					addStr(str, buildPathExpression(previousElement, pathEx));
					// addStr(str, "\t} \n";

				}

			}
		}else {

			if (previousElement.getType().equals("ForwardWildcard")) {
				addStr(str, "\t");
				addStr(str, Sparql.convertQueryElementToVariableName(previousElement));
				addStr(str, " ?p ?o . \n");
			}

			addStr(str, buildFilter(previousElement));
		}

		// remove duplicate lines
		// TODO: Make it cleverer. Currently, parentheses level is not considered.
		String[] arr = str.toString().split("\n");
		List<String> list = Arrays.asList(arr);
		Collections.reverse(list);
		arr = (String[]) list.toArray();

		ArrayList<String> result = new ArrayList<String>();
		String currentValue;
		for (int i = 0; i < arr.length; i++) {

			currentValue = arr[i];

			if (result.indexOf(currentValue) < 0 ||
					currentValue.trim().equals("FILTER (") ||
					currentValue.trim().equals(") .")) {
				result.add(currentValue);
			}

			// remove empty filters
			if (result.size() > 1 &&
					result.get(result.size() - 1).trim().equals("FILTER (") &&
					result.get(result.size() - 2).trim().equals(") .")) {

				result.remove(result.size() - 2);
				result.remove(result.size() - 1);
			}

		}

		arr = result.toArray(new String[0]);

		int index = -1;
		for (int l = 0; l < arr.length; l++) {
			index = arr[l].indexOf("##");
			if (index > -1) {
				arr[l] = arr[l].substring(0, index);
			}
		}

		list = Arrays.asList(arr);

		Collections.reverse(list);
		String res = "";
		for (String s : list) {
			res += s + "\n";
		}

		return res;

	};

	public static String buildFirstTriple(ResourceOrProperty elem1, ResourceOrProperty elem2) {

		StringBuilder str = new StringBuilder("\t");

		String value1 = elem1.getValue();

		// resource is wildcard
		if (value1.equals("*")) {
			value1 = Sparql.convertQueryElementToVariableName(elem1);
			// addStr(str, value1 + " a owl:Thing . \n\t";
		}

		String type = elem2.getType();

		switch(type) {
		case "ForwardProperty":
		case "ForwardPropertyWithFilter":
			addStr(str, value1 + " ");
			addStr(str, elem2.getValue() + " ");
			addStr(str, Sparql.convertQueryElementToVariableName(elem2) + " . \n");
			break;
		case "ReversedProperty":
		case "ReversedPropertyWithFilter":
			addStr(str, Sparql.convertQueryElementToVariableName(elem2) + " ");
			addStr(str, elem2.getValue() + " ");
			addStr(str, value1 + " . \n");
			break;
		case "ForwardWildcard":
		case "ForwardWildcardWithFilter":
			// String propetyName = Sparql.convertQueryElementToVariableName(elem1);
			addStr(str, value1 + " ");
			// addStr(str, "?PredicateFor" + propetyName.substr(1) + "__" + SemwidgQL.Sparql.convertQueryElementToVariableName(elem2).substr(1) + " ");
			addStr(str, "?PredicateFor" + Sparql.convertQueryElementToVariableName(elem2).substring(1) + " ");
			addStr(str, Sparql.convertQueryElementToVariableName(elem2) + " . \n");
			break;
		case "ReversedWildcard":
		case "ReversedWildcardWithFilter":
			String propetyName1 = Sparql.convertQueryElementToVariableName(elem2);
			addStr(str, propetyName1 + " ");
			// addStr(str, "?PredicateFor" + propetyName.substr(1) + "__" + SemwidgQL.Sparql.convertQueryElementToVariableName(elem1).substr(1) + " ");
			addStr(str, "?PredicateFor" + propetyName1.substring(1) + " ");
			addStr(str, value1 + " . \n");
			break;
		}

		return unescape(str.toString());
	};

	public static String buildTriple(ResourceOrProperty elem1, ResourceOrProperty elem2) {

		StringBuilder str = new StringBuilder("\t");
		String type = elem2.getType();

		switch(type) {
		case "ForwardProperty":
		case "ForwardPropertyWithFilter":
			addStr(str, Sparql.convertQueryElementToVariableName(elem1) + " ");
			addStr(str, elem2.getValue() + " ");
			addStr(str, Sparql.convertQueryElementToVariableName(elem2) + " .\n");
			break;
		case "ReversedProperty":
		case "ReversedPropertyWithFilter":
			addStr(str, Sparql.convertQueryElementToVariableName(elem2) + " ");
			addStr(str, elem2.getValue() + " ");
			addStr(str, Sparql.convertQueryElementToVariableName(elem1) + " .\n");
			break;
		case "ForwardWildcard":
		case "ForwardWildcardWithFilter":
			// String propetyName = Sparql.convertQueryElementToVariableName(elem1);
			addStr(str, Sparql.convertQueryElementToVariableName(elem1) + " ");
			// addStr(str, "?PredicateFor" + propetyName.substr(1) + "__" + SemwidgQL.Sparql.convertQueryElementToVariableName(elem2).substr(1) + " ");
			addStr(str, "?PredicateFor" + Sparql.convertQueryElementToVariableName(elem2).substring(1) + " ");
			addStr(str, Sparql.convertQueryElementToVariableName(elem2) + " .\n");
			break;
		case "ReversedWildcard":
		case "ReversedWildcardWithFilter":
			String propetyName1 = Sparql.convertQueryElementToVariableName(elem2);
			addStr(str, Sparql.convertQueryElementToVariableName(elem2) + " ");
			// addStr(str, "?PredicateFor" + propetyName.substr(1) + "__" + SemwidgQL.Sparql.convertQueryElementToVariableName(elem1).substr(1) + " ");
			addStr(str, "?PredicateFor" + propetyName1.substring(1) + " ");
			addStr(str, Sparql.convertQueryElementToVariableName(elem1) + " .\n");
			break;
		}

		return unescape(str.toString());
	};

	public static String buildFilter(ResourceOrProperty property) {

		StringBuilder str = new StringBuilder("");

		if (property.getType().equals("ForwardPropertyWithFilter") ||
				property.getType().equals("ReversedPropertyWithFilter") ||
				property.getType().equals("ForwardWildcardWithFilter") ||
				property.getType().equals("ReversedWildcardWithFilter") &&
				property.getFilter() != null) {

			ArrayList<Element> filter = property.getFilter();

			ArrayList<Element> realFilters = new ArrayList<Element>();
			ArrayList<Expression> pseudoFilters = new ArrayList<Expression>();

			// build triples to get values for Filter expression
			// and split into real and pseudo Filter expressions
			for (int i = 0; i < filter.size(); i = i + 2) {

				Expression filterExpression = (Expression) filter.get(i);
				addStr(str, buildFilterTriples(property, filterExpression));

				if (filterExpression instanceof KeywordExpression_Timeinterval ||
						filterExpression instanceof KeywordExpression_Hide ||
						filterExpression instanceof KeywordExpression_Aggregate ||
						filterExpression instanceof KeywordExpression_Optional ||
						filterExpression instanceof KeywordExpression_Predicate) {

					pseudoFilters.add(filterExpression);

				}else {

					if (realFilters.size() > 0 && i > 0) {
						realFilters.add(filter.get(i - 1));
					}
					realFilters.add(filterExpression);

				}
			}

			// build real Filter expression
			if (realFilters.size() > 0) {
				addStr(str, "\t");
				addStr(str, "FILTER (\n\t\t");

				for (int j = 0; j < realFilters.size(); j = j + 2) {

					Element filterExpression = realFilters.get(j);

					addStr(str, buildFilterExpression(property, filterExpression));

					if (j + 1 < realFilters.size() && realFilters.get(j + 1) instanceof ConditionalOperator) {
						addStr(str, " " + ((ConditionalOperator) realFilters.get(j + 1)).getValue() + "\n\t\t");
					}

				}

				addStr(str, "\n\t) .\n");

			}

			// build pseudo Filter expression
			for (int i = 0; i < pseudoFilters.size(); i++) {
				if (pseudoFilters.get(i) instanceof KeywordExpression_Timeinterval) {
					addStr(str, "\tBIND(FLOOR((xsd:dateTime(");
					addStr(str, Sparql.convertQueryElementToVariableName(property));
					addStr(str, ") - xsd:dateTime('1970-01-01T00:00:00'))");
					addStr(str, " / (" + Util.readPeriodExpression(((KeywordExpression_Timeinterval) pseudoFilters.get(i)).getTimeinterval()) + ") ) AS ");
					addStr(str, Sparql.convertQueryElementToVariableName(property) + "_timeinterval)\n");
				}
			}
		}

		return unescape(str.toString());

	};

	public static String buildFilterTriples(ResourceOrProperty property, Element filterExpression) {

		StringBuilder str = new StringBuilder("");

		if (filterExpression instanceof RelationalExpression) {

			ArrayList<ResourceOrProperty> properties = new ArrayList<ResourceOrProperty>();
			properties.add(property);
			properties.addAll(((RelationalExpression) filterExpression).getProperties());

			for (int j = 0; j < properties.size() - 1; j++) {
				addStr(str, buildTriple(properties.get(j), properties.get(j + 1)));
			}

			ArrayList<Element> values = ((RelationalExpression) filterExpression).getValue();

			for (int i = 0; i < values.size(); i++) {
				Element value = values.get(i);

				if (value instanceof SimpleQuery) {
					addStr(str, buildPathExpression(((SimpleQuery) value).getResource(), ((SimpleQuery) value).getPathExpression()));
				}
			}

		}
		else if (filterExpression instanceof KeywordExpression_Self) {

			ArrayList<Element> values = ((KeywordExpression_Self) filterExpression).getValue();

			for (int i = 0; i < values.size(); i++) {
				Element value = values.get(i);

				if (value instanceof SimpleQuery) {
					addStr(str, buildPathExpression(((SimpleQuery) value).getResource(), ((SimpleQuery) value).getPathExpression()));
				}
			}

		}
		else if (filterExpression instanceof KeywordExpression_Type) {

			String variableName = Sparql.convertQueryElementToVariableName(property);

			addStr(str, "\t");
			addStr(str, variableName);
			addStr(str, " rdf:type ");
			addStr(str, "?TypeOf_" + variableName.substring(1));
			addStr(str, " . \n");
		}

		return unescape(str.toString());

	};

	public static String buildFilterExpression(ResourceOrProperty property, Element filterExpression) {

		StringBuilder str = new StringBuilder("");

		if (filterExpression instanceof KeywordExpression_Language) {

			addStr(str, "lang(");
			addStr(str, Sparql.convertQueryElementToVariableName(property));
			addStr(str, ") = '' || ");
			addStr(str, "langMatches(lang(");
			addStr(str, Sparql.convertQueryElementToVariableName(property));
			addStr(str, "), ");
			addStr(str, ((KeywordExpression_Language) filterExpression).getLanguage());
			addStr(str, ")");

		}
		else if (filterExpression instanceof KeywordExpression_Type) {

			String variableName = Sparql.convertQueryElementToVariableName(property);
			addStr(str, "?TypeOf_" + variableName.substring(1));
			addStr(str, " " + ((KeywordExpression_Type) filterExpression).getOperator() + " ");
			addStr(str, ((KeywordExpression_Type) filterExpression).getType());

		}
		else if (filterExpression instanceof KeywordExpression_Self) {

			ArrayList<Element> values = ((KeywordExpression_Self) filterExpression).getValue();
			ArrayList<ArrayList<Property>> lastProperties = new ArrayList<ArrayList<Property>>();

			for (int i = 0; i < values.size(); i++) {
				Element value = values.get(i);

				if (value instanceof SimpleQuery) {

					if (((SimpleQuery) value).getPathExpression() != null) {
						lastProperties.add(getLastPropertiesOfPathExpression(((SimpleQuery) value).getPathExpression()));
					}

				}
			}

			if (lastProperties.size() == 0) {

				// 'CONTAINS' check
				if (((KeywordExpression_Self) filterExpression).getOperator().equals("~")) {
					addStr(str, "CONTAINS( ");
					addStr(str, "LCASE(STR(");
					addStr(str, Sparql.convertQueryElementToVariableName(property));
					addStr(str, ")), LCASE(STR(");

					for (int i = 0; i < values.size(); i++) {
						Element value = values.get(i);
						if (value instanceof SimpleQuery) {
							addStr(str, Sparql.convertQueryElementToVariableName(((SimpleQuery) value).getResource()));
						}else {
							addStr(str, value + " ");
						}
					}
					addStr(str, ")))");

				}else {
					// check if String
					if ((values.get(0).toString().charAt(0) + "").equals("'") || (values.get(0).toString().charAt(0) + "").equals("\"")) {
						addStr(str, "STR(");
						addStr(str, Sparql.convertQueryElementToVariableName(property));
						addStr(str, ") ");
					}else {
						addStr(str, Sparql.convertQueryElementToVariableName(property) + " ");
					}

					addStr(str, ((KeywordExpression_Self) filterExpression).getOperator() + " ");

					for (int i = 0; i < values.size(); i++) {
						Element value = values.get(i);
						if (value instanceof SimpleQuery) {
							addStr(str, ((SimpleQuery) value).getResource().getValue() + " ");
						}else {
							addStr(str, value + " ");
						}

					}
				}
			}else {
				int numberOfFilters = 1;

				for (int i = 0; i < lastProperties.size(); i++) {
					numberOfFilters *= lastProperties.get(i).size();
				}

				ArrayList<Integer> propertyIndices = new ArrayList<Integer>();
				int loops = lastProperties.size();
				for (int l = 0; l < loops; l++) {
					propertyIndices.add(0);
				}

				for (int j = 0; j < numberOfFilters; j++) {

					// 'CONTAINS' check
					if (((KeywordExpression_Self) filterExpression).getOperator().equals("~")) {
						addStr(str, "CONTAINS( ");
						addStr(str, "LCASE(STR(");

						addStr(str, Sparql.convertQueryElementToVariableName(property));
						addStr(str, ")), LCASE(");

						int simpleQueryIndex = 0;
						for (int k = 0; k < values.size(); k++) {
							Element value = values.get(k);

							if (value instanceof SimpleQuery) {
								ArrayList<Property> simpleQuerysProperties = lastProperties.get(simpleQueryIndex);
								Property simpleQuerysProperty = simpleQuerysProperties.get(propertyIndices.get(simpleQueryIndex));
								simpleQueryIndex++;
								addStr(str, Sparql.convertQueryElementToVariableName(simpleQuerysProperty) + " ");
							}else {
								addStr(str, value + " ");
							}
						}

						addStr(str, "))");

					}else {

						if ((values.get(0).toString().charAt(0) + "").equals("'") || (values.get(0).toString().charAt(0) + "").equals("\"")) {
							addStr(str, "STR(");
							addStr(str, Sparql.convertQueryElementToVariableName(property));
							addStr(str, ") ");
						}else {
							addStr(str, Sparql.convertQueryElementToVariableName(property) + " ");
						}

						addStr(str, ((KeywordExpression_Self) filterExpression).getOperator() + " ");

						int simpleQueryIndex = 0;
						for (int k = 0; k < values.size(); k++) {
							Element value = values.get(k);

							if (value instanceof SimpleQuery) {
								ArrayList<Property> simpleQuerysProperties = lastProperties.get(simpleQueryIndex);
								Property simpleQuerysProperty = simpleQuerysProperties.get(propertyIndices.get(simpleQueryIndex));
								simpleQueryIndex++;
								addStr(str, Sparql.convertQueryElementToVariableName(simpleQuerysProperty) + " ");
							}else {
								addStr(str, value + " ");
							}
						}
					}

					// Define last properties indices for next iteration
					boolean increaseNext = true;
					for (int m = 0; m < loops; m++) {

						int val = propertyIndices.get(m);
						int length = lastProperties.get(m).size();

						if (increaseNext){
							propertyIndices.set(m, (val + 1) % length);

							if ((val + 1) == propertyIndices.get(m)) {
								increaseNext = false;
							}else {
								increaseNext = true;
							}
						}
					}

					if (j < numberOfFilters - 1) {
						addStr(str, " || \n\t\t");
					}
				}
			}

		}
		else if (filterExpression instanceof RelationalExpression) {

			ArrayList<Element> values = ((RelationalExpression) filterExpression).getValue();
			ArrayList<ArrayList<Property>> lastProperties = new ArrayList<ArrayList<Property>>();

			for (int i = 0; i < values.size(); i++) {
				Element value = values.get(i);

				if (value instanceof SimpleQuery) {

					if (((SimpleQuery) value).getPathExpression() != null) {
						lastProperties.add(getLastPropertiesOfPathExpression(((SimpleQuery) value).getPathExpression()));
					}

				}
			}

			ArrayList<Property> properties = ((RelationalExpression) filterExpression).getProperties();

			if (lastProperties.size() == 0) {

				// 'CONTAINS' check
				if (((RelationalExpression) filterExpression).getOperator().equals("~")) {
					addStr(str, "CONTAINS( ");
					addStr(str, "LCASE(STR(");
					addStr(str, Sparql.convertQueryElementToVariableName(properties.get(properties.size() - 1)));
					addStr(str, ")), LCASE(STR(");

					for (int i = 0; i < values.size(); i++) {
						Element value = values.get(i);
						if (value instanceof SimpleQuery) {
							addStr(str, Sparql.convertQueryElementToVariableName(((SimpleQuery) value).getResource()));
						}else {
							addStr(str, value + " ");
						}
					}
					addStr(str, ")))");

				}else {
					// check if String
					if ((values.get(0).toString().charAt(0) + "").equals("'") || (values.get(0).toString().charAt(0) + "").equals("\"")) {
						addStr(str, "STR(");
						addStr(str, Sparql.convertQueryElementToVariableName(properties.get(properties.size() - 1)));
						addStr(str, ") ");
					}else {
						addStr(str, Sparql.convertQueryElementToVariableName(properties.get(properties.size() - 1)) + " ");
					}

					addStr(str, ((RelationalExpression) filterExpression).getOperator() + " ");

					for (int i = 0; i < values.size(); i++) {
						Element value = values.get(i);
						if (value instanceof SimpleQuery) {
							addStr(str, ((SimpleQuery) value).getResource().getValue() + " ");
						}else {
							addStr(str, value + " ");
						}

					}
				}
			}else {
				int numberOfFilters = 1;

				for (int i = 0; i < lastProperties.size(); i++) {
					numberOfFilters *= lastProperties.get(i).size();
				}

				ArrayList<Integer> propertyIndices = new ArrayList<Integer>();
				int loops = lastProperties.size();
				for (int l = 0; l < loops; l++) {
					propertyIndices.add(0);
				}

				for (int j = 0; j < numberOfFilters; j++) {

					// 'CONTAINS' check
					if (((RelationalExpression) filterExpression).getOperator().equals("~")) {
						addStr(str, "CONTAINS( ");
						addStr(str, "LCASE(STR(");

						addStr(str, Sparql.convertQueryElementToVariableName(properties.get(properties.size() - 1)));
						addStr(str, ")), LCASE(");

						int simpleQueryIndex = 0;
						for (int k = 0; k < values.size(); k++) {
							Element value = values.get(k);

							if (value instanceof SimpleQuery) {
								ArrayList<Property> simpleQuerysProperties = lastProperties.get(simpleQueryIndex);
								Property simpleQuerysProperty = simpleQuerysProperties.get(propertyIndices.get(simpleQueryIndex));
								simpleQueryIndex++;
								addStr(str, Sparql.convertQueryElementToVariableName(simpleQuerysProperty) + " ");
							}else {
								addStr(str, value + " ");
							}
						}

						addStr(str, "))");

					}else {

						if ((values.get(0).toString().charAt(0) + "").equals("'") || (values.get(0).toString().charAt(0) + "").equals("\"")) {
							addStr(str, "STR(");
							addStr(str, Sparql.convertQueryElementToVariableName(properties.get(properties.size() - 1)));
							addStr(str, ") ");
						}else {
							addStr(str, Sparql.convertQueryElementToVariableName(properties.get(properties.size() - 1)) + " ");
						}

						addStr(str, ((RelationalExpression) filterExpression).getOperator() + " ");

						int simpleQueryIndex = 0;
						for (int k = 0; k < values.size(); k++) {
							Element value = values.get(k);

							if (value instanceof SimpleQuery) {
								ArrayList<Property> simpleQuerysProperties = lastProperties.get(simpleQueryIndex);
								Property simpleQuerysProperty = simpleQuerysProperties.get(propertyIndices.get(simpleQueryIndex));
								simpleQueryIndex++;
								addStr(str, Sparql.convertQueryElementToVariableName(simpleQuerysProperty) + " ");
							}else {
								addStr(str, value + " ");
							}
						}
					}

					// Define last properties indices for next iteration
					boolean increaseNext = true;
					for (int m = 0; m < loops; m++) {

						int val = propertyIndices.get(m);
						int length = lastProperties.get(m).size();

						if (increaseNext){
							propertyIndices.set(m, (val + 1) % length);

							if ((val + 1) == propertyIndices.get(m)) {
								increaseNext = false;
							}else {
								increaseNext = true;
							}
						}
					}

					if (j < numberOfFilters - 1) {
						addStr(str, " || \n\t\t");
					}
				}
			}
		}
		else if (filterExpression instanceof KeywordExpression_Timestart) {
			addStr(str, "xsd:dateTime(" + Sparql.convertQueryElementToVariableName(property) + ")");
			addStr(str, " >= ");
			addStr(str, Util.readDateExpression(((KeywordExpression_Timestart) filterExpression).getTimestart()));
		}
		else if (filterExpression instanceof KeywordExpression_Timeend) {
			addStr(str, "xsd:dateTime(" + Sparql.convertQueryElementToVariableName(property) + ")");
			addStr(str, " <= ");
			addStr(str, Util.readDateExpression(((KeywordExpression_Timeend) filterExpression).getTimeend()));
		}

		return unescape(str.toString());

	};

	public static ArrayList<Property> getLastPropertiesOfPathExpression(PathExpression pathExpression) {

		ArrayList<Property> lastProperties = new ArrayList<Property>();

		if (pathExpression == null) {
			return lastProperties;
		}

		if (pathExpression.getMultiplePathExpression() != null) {
			ArrayList<PathExpression> multiplePathExpressions = pathExpression.getMultiplePathExpression().getPathExpressions();
			for (int i = 0; i < multiplePathExpressions.size(); i++) {
				PathExpression multiplePathExpression = multiplePathExpressions.get(i);
				lastProperties.addAll(getLastPropertiesOfPathExpression(multiplePathExpression));
			}
		}else if (pathExpression.getProperties() != null && pathExpression.getProperties().size() > 0){
			lastProperties.add(pathExpression.getProperties().get(pathExpression.getProperties().size() - 1));
		}

		return lastProperties;

	};

	public static String buildGroupPart(SimpleQuery simpleQuery) {

		StringBuilder str = new StringBuilder("");

		ArrayList<GroupExpression> groupExpressions = getGroupExpressions(simpleQuery);

		if (groupExpressions.size() > 0) {
			GroupExpression groupExpression;

			for (int i = 0; i < groupExpressions.size(); i++) {
				groupExpression = groupExpressions.get(i);

				if (groupExpression.getFilter() instanceof KeywordExpression_Timeinterval) {

					String name = Sparql.convertQueryElementToVariableName(groupExpression.getElement()) + "_timeinterval";

					addStr(str, "\nGROUP BY " + name);
					addStr(str, "\nORDER BY DESC(" + name + ")");
				}
			}
		}

		return str.toString();
	};

	public static ArrayList<GroupExpression> getGroupExpressions(Element elem) {

		ArrayList<GroupExpression> arr = new ArrayList<GroupExpression>();

		if (elem == null) {
			return arr;
		}

		int i = 0;

		if (elem instanceof SimpleQuery) {
			arr.addAll(getGroupExpressions(((SimpleQuery) elem).getPathExpression()));
		}
		else if (elem instanceof PathExpression) {
			if (((PathExpression) elem).getMultiplePathExpression() != null) {
				for (i = 0; i < ((PathExpression) elem).getMultiplePathExpression().getPathExpressions().size(); i++) {
					arr.addAll(getGroupExpressions(((PathExpression) elem).getMultiplePathExpression().getPathExpressions().get(i)));
				}
			}

			if (((PathExpression) elem).getProperties() != null) {
				for (i = 0; i < ((PathExpression) elem).getProperties().size(); i++) {
					arr.addAll(getGroupExpressions(((PathExpression) elem).getProperties().get(i)));
				}
			}
		}
		else if (elem instanceof Property) {
			if (((Property) elem).getFilter() != null) {
				for (i = 0; i < ((Property) elem).getFilter().size(); i++) {

					Element filter = ((Property) elem).getFilter().get(i);
					if (filter instanceof KeywordExpression_Timeinterval) {
						arr.add(new GroupExpression((Property) elem, (KeywordExpression_Timeinterval) filter));
					}
				}
			}
		}

		return arr;
	};

	public static void addStr(StringBuilder currentString, String toAdd) {
		currentString.append(toAdd);
	}

	public static void addStr(StringBuilder currentString, String toAdd, boolean debug) {
		System.out.println("~ " + currentString.toString() + " ~~ " + toAdd + " ~.");
		addStr(currentString, toAdd);
	}
}
