package org.semwidg.semwidgql.internal;

import java.util.ArrayList;

import org.antlr.runtime.ANTLRStringStream;
import org.antlr.runtime.CommonTokenStream;
import org.antlr.runtime.tree.CommonTree;
import org.semwidg.semwidgql.NamedResource;
import org.semwidg.semwidgql.Prefix;
import org.semwidg.semwidgql.internal.element.Resource;
import org.semwidg.semwidgql.internal.element.ResourceOrProperty;
import org.semwidg.semwidgql.internal.element.SimpleQuery;

public class SemwidgQLInternal {
	
	public static ArrayList<String> toSparql(String semwidgQuery, ArrayList<Prefix> prefixes, ArrayList<NamedResource> namedResources) {
	
		try {
			return toSparql(semwidgQuery, prefixes, namedResources, false);
		} catch (Exception e) {

		}
		
		return null;
	
	}
	
	public static ArrayList<String> toSparql(String semwidgQuery, ArrayList<Prefix> prefixes, ArrayList<NamedResource> namedResources, boolean throwExceptions) throws Exception {

		SemwidgQLInternal.setCurrentModel(null);

		if (semwidgQuery != null) {
			semwidgQuery = semwidgQuery.trim();
		}

		try {

			ANTLRStringStream cstream = new ANTLRStringStream(semwidgQuery);
			SemwidgQLLexer lexer = new SemwidgQLLexer(cstream);
			CommonTokenStream tstream = new CommonTokenStream(lexer);
			SemwidgQLParser parser = new SemwidgQLParser(tstream);
			
			CommonTree ast = (CommonTree) parser.semwidg().getTree();

			SemwidgQLInternal.setCurrentModel(new Model());
			SemwidgQLInternal.getCurrentModel().setPrefixList(prefixes);
			SemwidgQLInternal.getCurrentModel().setNamedResourceList(namedResources);

			ArrayList<String> sparqlQueries = Sparql.createSparqlQueries(ast);

			return sparqlQueries;

		}catch(Exception e) {
			if (throwExceptions) {
				throw e;
			}else {
				e.printStackTrace();
			}
		}

		return null;
	};

	public static ArrayList<String> getNamedResources(String semwidgQuery, boolean throwExceptions) throws Exception {
		
		ArrayList<String> namedResources = new ArrayList<String>();
		
		if (semwidgQuery != null) {
			semwidgQuery = semwidgQuery.trim();
		}
		
		try {

			ANTLRStringStream cstream = new ANTLRStringStream(semwidgQuery);
			SemwidgQLLexer lexer = new SemwidgQLLexer(cstream);
			CommonTokenStream tstream = new CommonTokenStream(lexer);
			SemwidgQLParser parser = new SemwidgQLParser(tstream);
			
			CommonTree ast = (CommonTree) parser.semwidg().getTree();

			SemwidgQLInternal.setCurrentModel(new Model());
			
			ArrayList<SimpleQuery> simpleQueries = Sparql.getSimpleQueries(ast);

			ArrayList<ResourceOrProperty> resAndProp = new ArrayList<ResourceOrProperty>();
			for (int i = 0; i < simpleQueries.size(); i++) {
				resAndProp.addAll(Sparql.getUsedResourcesAndProperties(simpleQueries.get(i)));
			}

			for (int i = 0; i < resAndProp.size(); i++) {
				if (resAndProp.get(i) instanceof Resource && ((Resource) resAndProp.get(i)).getType().equals("NamedResource")) {
					namedResources.add(((Resource) resAndProp.get(i)).getValue());
				}
			}
			
		}catch(Exception e) {
			if (throwExceptions) {
				throw e;
			}else {
				e.printStackTrace();
			}
		}
		
		return namedResources;
	}
	
	private static Model currentModel = null;

	public static Model getCurrentModel() {
		if (currentModel == null) {
			currentModel = new Model();
		}
		return currentModel;
	}

	public static void setCurrentModel(Model currentModel) {
		SemwidgQLInternal.currentModel = currentModel;
	}
	
	
}
