package org.semwidg.semwidgql.internal;

import java.util.ArrayList;

import org.antlr.runtime.tree.CommonTree;
import org.semwidg.semwidgql.NamedResource;
import org.semwidg.semwidgql.Prefix;
import org.semwidg.semwidgql.internal.element.ConditionalOperator;
import org.semwidg.semwidgql.internal.element.Element;
import org.semwidg.semwidgql.internal.element.Expression;
import org.semwidg.semwidgql.internal.element.KeywordExpression;
import org.semwidg.semwidgql.internal.element.KeywordExpression_Aggregate;
import org.semwidg.semwidgql.internal.element.KeywordExpression_Hide;
import org.semwidg.semwidgql.internal.element.KeywordExpression_Language;
import org.semwidg.semwidgql.internal.element.KeywordExpression_Optional;
import org.semwidg.semwidgql.internal.element.KeywordExpression_Predicate;
import org.semwidg.semwidgql.internal.element.KeywordExpression_Self;
import org.semwidg.semwidgql.internal.element.KeywordExpression_Timeend;
import org.semwidg.semwidgql.internal.element.KeywordExpression_Timeinterval;
import org.semwidg.semwidgql.internal.element.KeywordExpression_Timestart;
import org.semwidg.semwidgql.internal.element.KeywordExpression_Type;
import org.semwidg.semwidgql.internal.element.MultiplePathExpression;
import org.semwidg.semwidgql.internal.element.PathExpression;
import org.semwidg.semwidgql.internal.element.Property;
import org.semwidg.semwidgql.internal.element.RelationalExpression;
import org.semwidg.semwidgql.internal.element.Resource;
import org.semwidg.semwidgql.internal.element.ResourceOrProperty;
import org.semwidg.semwidgql.internal.element.SimpleQuery;
import org.semwidg.semwidgql.internal.element.StringElement;

public class Sparql {

	public static ArrayList<String> createSparqlQueries(CommonTree ast) throws Exception {

		ArrayList<String> sparqlQueries = new ArrayList<String>();
		ArrayList<SimpleQuery> simpleQueries = getSimpleQueries(ast);

		applyUrisToNamedResources();

		if (simpleQueries != null) {

			for (int i = 0; i < simpleQueries.size(); i++) {

				SimpleQuery simpleQuery = simpleQueries.get(i);

				String sparqlQuery = Factory.buildQuery(simpleQuery);

				if (sparqlQuery != null) {
					sparqlQueries.add(sparqlQuery);
				}
			}
		}

		return sparqlQueries;

	};

	public static ArrayList<String> getUsedPrefixStrings(CommonTree node, ArrayList<String> prefixes) {

		if (node == null) {
			return null;
		}

		if (prefixes == null) {
			prefixes = new ArrayList<String>();
		}

		if (node.getChildren() != null) {
			for (int i = 0; i < node.getChildren().size(); i++) {
				getUsedPrefixStrings((CommonTree) node.getChild(i), prefixes);
			}
		}

		if (Util.checkAstType(node, "NamespacePrefix")) {
			String prefix = node.getText();
			if (prefix.charAt(prefix.length() - 1) == ':') {
				prefix = prefix.substring(0, prefix.length() - 1);
			}
			boolean contains = false;
			for (int i = 0; i < prefixes.size(); i++) {
				if (prefixes.get(i).equals(prefix)) {
					contains = true;
					break;
				}
			}
			if (!contains) {
				prefixes.add(prefix);
			}
		}

		return prefixes;

	};

	public static ArrayList<Prefix> getUsedPrefixObjects(ArrayList<String> prefixStrings) {

		ArrayList<Prefix> pre = new ArrayList<Prefix>();
		ArrayList<Prefix> prefixList = SemwidgQLInternal.getCurrentModel().getPrefixList();

		if (prefixStrings != null && prefixStrings.size() > 0 && prefixList != null && prefixList.size() > 0) {
			for (int i = 0; i < prefixStrings.size(); i++) {

				String prefixString = prefixStrings.get(i);

				for (int j = 0; j < prefixList.size(); j++) {

					Prefix prefixObject = prefixList.get(j);

					if (prefixString.equals(prefixObject.getPrefix())) {
						pre.add(prefixObject);
						break;
					}
				}
			}
		}

		return pre;

	};

	public static ArrayList<ResourceOrProperty> getUsedResourcesAndProperties(SimpleQuery simpleQuery) {

		if (simpleQuery == null) {
			return null;
		}

		ArrayList<ResourceOrProperty> elements = new ArrayList<ResourceOrProperty>();
		Resource resource = simpleQuery.getResource();
		elements.add(resource);

		if (resource.getFilter() != null) {
			ArrayList<Element> value;
			for (int i = 0; i < resource.getFilter().size(); i++) {
				if (resource.getFilter().get(i) instanceof RelationalExpression) {
					value = ((RelationalExpression) resource.getFilter().get(i)).getValue();
					if (value != null && value.get(0) instanceof SimpleQuery) {
						elements.addAll(getUsedResourcesAndProperties((SimpleQuery) value.get(0)));
					}
				}
			}
		}


		PathExpression pathExpression = simpleQuery.getPathExpression();
		elements.addAll(getUsesResourcesAndPropertiesFromPathExpression(pathExpression));

		return elements;

	};

	private static ArrayList<ResourceOrProperty> getUsesResourcesAndPropertiesFromPathExpression(PathExpression pathExpression) {

		ArrayList<ResourceOrProperty> elements = new ArrayList<ResourceOrProperty>();

		if (pathExpression != null) {

			ArrayList<Property> properties = pathExpression.getProperties();
			if (properties != null) {
				for (int i = 0; i < properties.size(); i++) {
					Property property = properties.get(i);
					elements.add(property);

					if (property.getType().equals("ForwardPropertyWithFilter") ||
							property.getType().equals("ReversedPropertyWithFilter")) {

						ArrayList<Element> filters = property.getFilter();
						for (int j = 0; j < filters.size(); j++) {
							Element filter = filters.get(j);
							if (filter instanceof RelationalExpression) {
								ArrayList<Element> values = ((RelationalExpression) filter).getValue();
								for (int k = 0; k < values.size(); k++) {
									Element value = values.get(k);
									if (value instanceof SimpleQuery) {
										elements.addAll(getUsedResourcesAndProperties((SimpleQuery) value));
									}
								}
							}
						}
					}
				}
			}

			MultiplePathExpression multiplePathExpression = pathExpression.getMultiplePathExpression();
			if (multiplePathExpression != null && multiplePathExpression.getPathExpressions() != null && multiplePathExpression.getPathExpressions().size() > 0) {
				for (int j = 0; j < multiplePathExpression.getPathExpressions().size(); j++) {
					PathExpression multiPathEx = multiplePathExpression.getPathExpressions().get(j);
					elements.addAll(getUsesResourcesAndPropertiesFromPathExpression(multiPathEx));
				}
			}
		}

		return elements;
	};

	public static ArrayList<SimpleQuery> getSimpleQueries(CommonTree node) {

		ArrayList<SimpleQuery> simpleQueries = new ArrayList<SimpleQuery>();

		if (node == null) {
			return simpleQueries;
		}

		if (!Util.checkAstType(node, "SimpleQuery")) {

			if (!node.getChildren().isEmpty()) {
				for (int i = 0; i < node.getChildren().size(); i++) {
					simpleQueries.addAll(getSimpleQueries((CommonTree) node.getChild(i)));
				}
			}

		}else {

			SimpleQuery simpleQuery = new SimpleQuery();
			simpleQuery.setPointer(node);

			if (node.getChildren() != null && node.getChildren().size() >= 1 && node.getChild(0) != null) {
				Resource resource = getResource((CommonTree) node.getChild(0));
				simpleQuery.setResource(resource);
			}

			if (node.getChildren() != null && node.getChildren().size() >= 2 && node.getChild(1) != null) {
				PathExpression pathExpression = getPathExpression((CommonTree) node.getChild(1));
				simpleQuery.setPathExpression(pathExpression);
			}

			simpleQueries.add(simpleQuery);

		}

		return simpleQueries;

	};

	public static Resource getResource (CommonTree node) {

		if (node == null) {
			return null;
		}

		if (Util.checkAstType(node, "Resource")) {

			CommonTree res = (CommonTree) node.getChild(0);
			String type = Util.getTypeLabel(res);
			String value = null;
			ArrayList<Element> filter = null;

			switch (type) {
			case "NamedResource":
				value = res.getChild(0).getText();
				break;
			case "PrefixedResource":
				value = res.getChild(0).getText();
				value += res.getChild(1).getText();
				break;
			case "QualifiedResource":
				value = "<" + res.getChild(0).getText() + ">";
				break;
			case "ForwardWildcard":
				value = "*";
				break;
			case "ForwardWildcardWithFilter":
				value = "*";
				filter = getFilter((CommonTree) res.getChild(0));
				break;
			default:
				System.err.println("Type '" + type + "' for a Resource is unknown.");
				break;
			}

			if (value != null) {
				Resource resource = new Resource();
				resource.setType(type);
				resource.setValue(value);
				resource.setPointer(node);
				resource.setFilter(filter);

				SemwidgQLInternal.getCurrentModel().addToAllResourcesList(resource);

				return resource;
			}

		}else {
			System.err.println("Expected element of type 'Resource'. Instead, element of type '" + Util.getTypeLabel((CommonTree) node.getChild(0)) + "' was commited.");
		}

		return null;
	};

	public static PathExpression getPathExpression(CommonTree node) {

		if (node == null) {
			return null;
		}

		if (Util.checkAstType(node, "PathExpression")) {

			ArrayList<Property> properties = getProperties(node);
			MultiplePathExpression multiplePathExpression = getMultiplePathExpression(node);

			if (properties != null || multiplePathExpression != null) {
				PathExpression pathExpression = new PathExpression();
				pathExpression.setProperties(properties);
				pathExpression.setMultiplePathExpression(multiplePathExpression);
				pathExpression.setPointer(node);

				return pathExpression;
			}

		}else {
			System.err.println("Expected element of type 'PathExpression'. Instead, element of type '" + Util.getTypeLabel((CommonTree) node.getChild(0)) + "' was commited.");
		}

		return null;
	};

	public static ArrayList<Property> getProperties(CommonTree node) {

		if (node == null || node.getChildren() == null) {
			return null;
		}

		if (Util.checkAstType(node, "PathExpression") || Util.checkAstType(node, "RelationalExpression")) {

			if (node.getChildren() != null) {

				@SuppressWarnings("unchecked")
				ArrayList<CommonTree> children = (ArrayList<CommonTree>) node.getChildren();
				ArrayList<Property> properties = new ArrayList<Property>();

				for (int i = 0; i < children.size(); i++) {
					if (Util.checkAstType(children.get(i), "Property")) {

						CommonTree prop = (CommonTree) children.get(i).getChild(0);
						String type = Util.getTypeLabel(prop);
						String value = null;
						ArrayList<Element> filter = null;

						switch (type) {
						case "ForwardProperty":
							value = Util.getPropertyValue((CommonTree) prop.getChild(0), false);
							break;
						case "ForwardPropertyWithFilter":
							value = Util.getPropertyValue((CommonTree) prop.getChild(0), false);
							filter = getFilter((CommonTree) prop.getChild(1));
							break;
						case "ReversedProperty":
							value = Util.getPropertyValue((CommonTree) prop.getChild(0), false);
							break;
						case "ReversedPropertyWithFilter":
							value = Util.getPropertyValue((CommonTree) prop.getChild(0), false);
							filter = getFilter((CommonTree) prop.getChild(1));
							break;
						case "ForwardWildcard":
							value = "*";
							break;
						case "ForwardWildcardWithFilter":
							value = "*";
							filter = getFilter((CommonTree) prop.getChild(0));
							break;
						case "ReversedWildcard":
							value = "*";
							break;
						case "ReversedWildcardWithFilter":
							value = "*";
							filter = getFilter((CommonTree) prop.getChild(0));
							break;
						default:
							System.err.println("Type '" + type + "' for a Property is unknown.");
							break;
						}

						if (value != null) {
							Property property = new Property();
							property.setType(type);
							property.setValue(value);
							property.setFilter(filter);
							property.setPointer(prop);

							SemwidgQLInternal.getCurrentModel().addToAllPropertiesList(property);

							properties.add(property);
						}
					}
				}

				if (properties.size() > 0) {
					return properties;
				}

			}

		}else {
			System.err.println("Expected element of type 'PathExpression'. Instead, element of type '" + Util.getTypeLabel(node) + "' was commited.");
		}

		return null;

	};

	public static ArrayList<Element> getFilter(CommonTree node) {

		if (node == null || node.getChildren() == null) {
			return null;
		}

		ArrayList<Element> filter = null;

		if (Util.checkAstType(node, "ConditionalExpression")) {

			filter = new ArrayList<Element>();

			CommonTree orExpressions = (CommonTree) node.getChild(0);

			ConditionalOperator conditionalAnd = new ConditionalOperator();
			conditionalAnd.setValue("&&");

			ConditionalOperator conditionalOr = new ConditionalOperator();
			conditionalOr.setValue("||");

			for (int i = 0; i < orExpressions.getChildren().size(); i++) {

				CommonTree andExpressions = (CommonTree) orExpressions.getChild(i);

				for (int j = 0; j < andExpressions.getChildren().size(); j++) {

					CommonTree expression = (CommonTree) andExpressions.getChild(j);

					filter.add(getExpression(expression));

					if (j < andExpressions.getChildren().size() - 1) {
						filter.add(conditionalAnd);
					}
				}

				if (i < orExpressions.getChildren().size() - 1) {
					filter.add(conditionalOr);
				}
			}

		}

		return filter;
	};

	public static Expression getExpression(CommonTree node) {

		if (node == null || node.getChildren() == null) {
			return null;
		}

		Expression expression = null;

		if (Util.checkAstType(node, "Expression")) {

			String type = Util.getTypeLabel((CommonTree) node.getChild(0));

			switch (type) {
			case "KeywordExpression":
				expression = getKeywordExpression((CommonTree) node.getChild(0));
				break;
			case "RelationalExpression":
				expression = getRelationalExpression((CommonTree) node.getChild(0));
				break;
			}

		}

		return expression;
	};

	public static KeywordExpression getKeywordExpression(CommonTree node) {

		if (node == null || node.getChildren() == null) {
			return null;
		}

		KeywordExpression keywordExpression = null;

		if (Util.checkAstType(node, "KeywordExpression")) {

			keywordExpression = new KeywordExpression();

			String type = Util.getTypeLabel((CommonTree) node.getChild(0));

			switch (type) {
			case "Language":
				if (node.getChild(0).getChild(0) != null && node.getChild(0).getChild(0).getChild(0) != null) {

					keywordExpression = new KeywordExpression_Language();
					keywordExpression.setOperator("=");
					((KeywordExpression_Language) keywordExpression).setLanguage(node.getChild(0).getChild(0).getChild(0).getText());
				}
				break;
			case "Type":
				if (node.getChild(0).getChild(0)  != null&& node.getChild(0).getChild(0).getChild(0) != null) {

					keywordExpression = new KeywordExpression_Type();
					keywordExpression.setOperator(node.getChild(0).getChild(0).getChild(0).getText());
					((KeywordExpression_Type) keywordExpression).setType(Util.getPropertyValue((CommonTree) node.getChild(0).getChild(1).getChild(0), false));
				}
				break;
			case "Self":
				if (node.getChild(0).getChild(0)  != null&& node.getChild(0).getChild(0).getChild(0) != null 
				&& node.getChild(0).getChild(1) != null && node.getChild(0).getChild(1).getChild(0) != null) {

					keywordExpression = new KeywordExpression_Self();
					keywordExpression.setOperator(node.getChild(0).getChild(0).getChild(0).getText());
					((KeywordExpression_Self) keywordExpression).setValue(getRelationalExpressionValue((CommonTree) node.getChild(0).getChild(1).getChild(0)));
				}
				break;
			case "Timeinterval":
				if (node.getChild(0).getChild(0) != null && node.getChild(0).getChild(0).getChild(0) != null) {

					keywordExpression = new KeywordExpression_Timeinterval();
					keywordExpression.setOperator("=");

					String timeinterval = node.getChild(0).getChild(0).getChild(0).getText();
					timeinterval = timeinterval.substring(1, timeinterval.length() - 1);
					((KeywordExpression_Timeinterval) keywordExpression).setTimeinterval(timeinterval);
				}
				break;
			case "Timestart":
				if (node.getChild(0).getChild(0) != null && node.getChild(0).getChild(0).getChild(0) != null) {

					keywordExpression = new KeywordExpression_Timestart();
					keywordExpression.setOperator("=");

					String timestart = node.getChild(0).getChild(0).getChild(0).getText();
					timestart = timestart.substring(1, timestart.length() - 1);
					((KeywordExpression_Timestart) keywordExpression).setTimestart(timestart);
				}
				break;
			case "Timeend":
				if (node.getChild(0).getChild(0) != null && node.getChild(0).getChild(0).getChild(0) != null) {

					keywordExpression = new KeywordExpression_Timeend();
					keywordExpression.setOperator("=");

					String timeend = node.getChild(0).getChild(0).getChild(0).getText();
					timeend = timeend.substring(1, timeend.length() - 1);
					((KeywordExpression_Timeend) keywordExpression).setTimeend(timeend);
				}
				break;
			case "Aggregate":
				if (node.getChild(0).getChild(0) != null && node.getChild(0).getChild(0).getChild(0) != null) {

					keywordExpression = new KeywordExpression_Aggregate();
					keywordExpression.setOperator("=");

					String aggregate = node.getChild(0).getChild(0).getChild(0).getText();
					aggregate = aggregate.substring(1, aggregate.length() - 1);
					((KeywordExpression_Aggregate) keywordExpression).setAggregate(aggregate);
				}
				break;
			case "Hide":
				if (node.getChild(0).getChild(0) != null && node.getChild(0).getChild(0).getChild(0) != null) {

					keywordExpression = new KeywordExpression_Hide();
					keywordExpression.setOperator("=");

					String hide = node.getChild(0).getChild(0).getChild(0).getText();
					((KeywordExpression_Hide) keywordExpression).setHide(hide);
				}
				break;
			case "Optional":
				if (node.getChild(0).getChild(0) != null && node.getChild(0).getChild(0).getChild(0) != null) {

					keywordExpression = new KeywordExpression_Optional();
					keywordExpression.setOperator("=");

					String optional = node.getChild(0).getChild(0).getChild(0).getText();
					((KeywordExpression_Optional) keywordExpression).setOptional(optional);
				}
				break;
			case "Predicate":
				if (node.getChild(0).getChild(0) != null && node.getChild(0).getChild(0).getChild(0) != null) {

					keywordExpression = new KeywordExpression_Predicate();
					keywordExpression.setOperator("=");

					String predicate = node.getChild(0).getChild(0).getChild(0).getText();
					((KeywordExpression_Predicate) keywordExpression).setPredicate(predicate);
				}
				break;
			default:
				System.err.println("Type '" + type + "' for a KeywordExpression is unknown.");
			}
		}

		return keywordExpression;

	};

	public static RelationalExpression getRelationalExpression(CommonTree node) {

		if (node == null || node.getChildren() == null) {
			return null;
		}

		RelationalExpression relationalExpression = null;

		if (Util.checkAstType(node, "RelationalExpression")) {

			relationalExpression = new RelationalExpression();

			ArrayList<Property> properties = getProperties(node);
			relationalExpression.setProperties(properties);

			String operator = node.getChild(node.getChildCount() - 2).getChild(0).getText();
			relationalExpression.setOperator(operator);

			ArrayList<Element> value = getRelationalExpressionValue((CommonTree) node.getChild(node.getChildCount() - 1));
			relationalExpression.setValue(value);
		}

		return relationalExpression;
	};

	public static ArrayList<Element> getRelationalExpressionValue(CommonTree node) {

		ArrayList<Element> values = new ArrayList<Element>();

		values.addAll(getRelationalExpressionValueRecursive(node));

		return values;


	};

	private static ArrayList<Element> getRelationalExpressionValueRecursive(CommonTree node) {

		ArrayList<Element> values = new ArrayList<Element>();

		String type = Util.getTypeLabel(node);

		switch (type) {
		case "SimpleQuery":
			values.addAll(getSimpleQueries(node));
			break;
		case "QualifiedProperty":
		case "PrefixedProperty":
			StringElement str = new StringElement();
			for (int i = 0; i < node.getChildren().size(); i++) {
				str.setValue(str.getValue() + node.getChild(i).getText());
			}
			values.add(str);
			break;
		default:
			if (node.getChildren() != null && node.getChildren().size() > 0) {
				for (int i = 0; i < node.getChildren().size(); i++) {
					values.addAll(getRelationalExpressionValueRecursive((CommonTree) node.getChild(i)));
				}
			}else {
				StringElement str2 = new StringElement();
				str2.setValue(node.getText());
				values.add(str2);
			}
			break;
		}

		return values;
	};

	public static MultiplePathExpression getMultiplePathExpression(CommonTree node) {

		if (node == null || node.getChildren() == null) {
			return null;
		}

		if (Util.checkAstType(node, "PathExpression")) {

			CommonTree lastChild = (CommonTree) node.getChild(node.getChildren().size() - 1);
			MultiplePathExpression multiplePathExpression = null;

			if (Util.checkAstType(lastChild, "MultiplePathExpression") && lastChild.getChildren() != null) {


				ArrayList<PathExpression> pathExpressions = new ArrayList<PathExpression>();

				for (int i = 0; i < lastChild.getChildren().size(); i++) {

					PathExpression pathExpression = getPathExpression((CommonTree) lastChild.getChild(i));

					if (pathExpression != null) {

						pathExpressions.add(pathExpression);

					}
				}

				if (pathExpressions.size() > 0) {
					multiplePathExpression = new MultiplePathExpression();
					multiplePathExpression.setPathExpressions(pathExpressions);
					multiplePathExpression.setPointer(node);

					return multiplePathExpression;
				}

			}



		}else {
			System.err.println("Expected element of type 'PathExpression'. Instead, element of type '" + Util.getTypeLabel(node) + "' was commited.");
		}

		return null;
	};

	public static void applyUrisToNamedResources() throws Exception {

		ArrayList<Resource> resources = SemwidgQLInternal.getCurrentModel().getAllResourcesList();
		ArrayList<NamedResource> namedResources = SemwidgQLInternal.getCurrentModel().getNamedResourceList();

		if (resources == null || namedResources == null) {
			return;
		}

		for (int i = 0; i < resources.size(); i++) {
			Resource res = resources.get(i);

			if (res.getType().equals("NamedResource")) {

				boolean resourceUriApplied = false;

				for (int j = 0; j < namedResources.size(); j++) {
					NamedResource named = namedResources.get(j);

					if (named.getName().equals(res.getValue())) {

						// URI
						if (named.getValue().indexOf("//") >= 0) {
							res.setValue("<" + named.getValue() + ">");
							res.setType("QualifiedResource");
							resourceUriApplied = true;
							// Prefixed Resource
						}else if (named.getValue().indexOf(":") >= 0) {
							res.setValue(named.getValue());
							res.setType("PrefixedResource");
							resourceUriApplied = true;
						}

						break;
					}
				}

				if (!resourceUriApplied) {
					throw new Exception("Parser Error! '" + res.getValue() + "' is not a valid resource name.");
				}
			}
		}
	};

	public static String convertQueryElementToVariableName(ResourceOrProperty elem) {

		String variableName = "";

		if (elem instanceof Resource) {

			if (elem.getValue().equals("*")) {
				variableName += "wildcard";
			}else {
				variableName += valueToLabel(elem.getValue());
			}
			variableName += "__R";
			variableName += treePathToLabelSuffix(elem.getPointer());

		}else if (elem instanceof Property) {

			if (elem.getValue().equals("*")) {
				variableName += "wildcard";
			}else {
				variableName += valueToLabel(elem.getValue());
			}

			if (elem.getType().equals("ReversedProperty") || elem.getType().equals("ReversedPropertyWithFilter")) {
				variableName += "Of";
			}

			variableName += "__P";
			variableName += treePathToLabelSuffix(elem.getPointer());

		}

		variableName = "?" + variableName.replaceAll("/([^a-zA-Z0-9_ ])|^[_0-9]+/g",  "_");

		return variableName;

	};

	public static String valueToLabel(String value) {

		String label = "";

		if (value.indexOf("<") == 0) {

			String uri = value.substring(1, value.length() - 1);

			// remove trailing slash
			if (uri.lastIndexOf("/") == uri.length() - 1) {
				uri = uri.substring(0, uri.length() - 1);
			}

			int slash = uri.lastIndexOf("/");
			int hash = uri.lastIndexOf("#");

			label = uri.substring(Math.max(slash, hash) + 1);

		}else {

			int colon = value.lastIndexOf(":");
			label = value.substring(colon + 1);
		}

		return label;
	};

	public static String treePathToLabelSuffix(CommonTree node) {

		String suffix = "";

		String type = Util.getTypeLabel(node);
		switch(type) {
		case "Resource":
		case "Property":
		case "PathExpression":
		case "MultiplePathExpression":
		case "SimpleQuery":
			suffix = "_" + node.getChildIndex();
			break;
		}

		if (node.getParent() == null) {
			return suffix;
		}else {
			return treePathToLabelSuffix((CommonTree) node.getParent()) + suffix;
		}

	};

}