package org.semwidg.semwidgql.internal;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.TimeZone;

import javax.xml.bind.DatatypeConverter;

import org.antlr.runtime.tree.CommonTree;
import org.semwidg.semwidgql.Prefix;

public class Util {

	public static String truncate(String str, int maxLength, String suffix) {

		if(str.length() > maxLength + suffix.length() + 1) {
			str = str.substring(0, maxLength + 1);
			if (str.lastIndexOf(" ") > 0) {
				str = str.substring(0, Math.min(str.length(), str.lastIndexOf(" ")));
			}
			str = str + suffix;
		}

		return str;
	};

	public static String getTypeLabel(CommonTree node) {
		if (node == null) {
			return null;
		}

		return SemwidgQLParser.tokenNames[node.getType()];
	};

	public static boolean checkAstType(CommonTree node, String type) {

		if (node == null || type == null || type.length() == 0) {
			return false;
		}

		return type.equals(getTypeLabel(node));
	};

	public static String getPropertyValue(CommonTree node, boolean removePefix) {
		if (node == null) {
			return null;
		}

		if (checkAstType(node, "PrefixedProperty")){

			if (removePefix) {
				String prefix = node.getChild(0).getText().substring(0, node.getChild(0).getText().length() - 1);
				ArrayList<Prefix> prefixList = SemwidgQLInternal.getCurrentModel().getPrefixList();
				for (int i = 0; i < prefixList.size(); i++) {
					String p = prefixList.get(i).getPrefix();
					if (p == prefix) {
						return "<" + prefixList.get(i).getUrl() + node.getChild(1).getText() + ">";
					}
				}
			}

			return node.getChild(0).getText() + node.getChild(1).getText();
		}else if (checkAstType(node, "QualifiedProperty")) {
			return "<" + node.getChild(1).getText() + ">";
		}
		
		return null;

	};

	// empty expressions return "now()"
	public static String readDateExpression(String expressionString) {
		String str;
		
// TODO: check regex
		String[] expression = ("" + expressionString).replace("/\\s+/g", " ").trim().split(" ");
		
		String date = "now()";
		String operator = "-";
		int seconds = 0;

		String exp;
		
		int state = 0;

		for (int i = 0; i < expression.length; i++) {
			exp = expression[i].trim();
			
			if (exp != "") {
				// is date?
				if (state < 1 && exp.indexOf("-") > -1) {
					Calendar cal =  DatatypeConverter.parseDateTime(exp);
					Date tmp = cal.getTime();
					
					TimeZone tz = TimeZone.getTimeZone("UTC");
					DateFormat df = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.SSSXXX");
					df.setTimeZone(tz);
					
					date = "xsd:dateTime('" + df.format(tmp) + "')";
					state = 1;
					continue;
				}
				
				// is now?
				if (state < 1 && exp.toLowerCase().indexOf("now") == 0) {
					date = "now()";
					state = 1;
					continue;
				}

				// is - or +?
				if (state < 2 && exp == "-" || exp == "+") {
					operator = exp;
					state = 2;
					continue;
				}

				// is period part
				float number = -1;
				try {
					number = Float.parseFloat(exp);
				} catch (NumberFormatException e) {
					//e.printStackTrace();
				}
				
				if (number >= 0) {
					String unit = null;

					unit = "s";
					if (expression.length > i + 1) {
						String exp2 = expression[i + 1].trim();
						if (exp2.length() > 0) {
							if (getMultiplicator(("" + exp2.charAt(0)).toLowerCase() ) > 0) {
								unit = ("" + exp2.charAt(0)).toLowerCase();
								i++;
							}
						}
					}
						

					int mul = getMultiplicator(unit);

					seconds += number * mul;

					state = 3;
				}
				
			}
		}

		str = date;

		if (seconds != 0) {
			str += " " + operator + " " + seconds;
		} 

		return str;
	};

	public static int readPeriodExpression(String expressionString) {

		String[] expression = ("" + expressionString).replace("/\\s+/g", " ").trim().split(" ");

		String exp;
		int seconds = 0;

		for (int i = 0; i < expression.length; i++) {
			exp = expression[i].trim();
			
			if (exp != "") {

				float number = -1;
				try {
					number = Float.parseFloat(exp);
				} catch (NumberFormatException e) {
//					e.printStackTrace();
				}

				if (number >= 0) {
					String unit = null;

					unit = "s";
					if (expression.length > i + 1) {
						String exp2 = expression[i + 1].trim();
						if (exp2.length() > 0) {
							if (getMultiplicator(("" + exp2.charAt(0)).toLowerCase()) > 0) {
								unit = ("" + exp2.charAt(0)).toLowerCase();
								i++;
							}
						}
					}
	
					int mul = getMultiplicator(unit);
	
					seconds += number * mul;
				}
			}
		}

		return seconds;
	};

	private static int getMultiplicator(String str) {
		
		int mul = 0;
		
		if (str != null) {
			switch(str) {
				case "s":
					mul = 1;
					break;
				case "m":
					mul = 60;
					break;
				case "h":
					mul = 3600;
					break;
				case "d":
					mul = 86400;
					break;
				case "w":
					mul = 604800;
					break;
				default:
					mul = 0;
					break;
			}
		}
		return mul;
	};
}
